#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <getopt.h>
#include <immintrin.h>
#include <inttypes.h>
#include <limits.h>
#include <linux/if_packet.h>
#include <math.h>
#include <netdb.h>
#include <netinet/ether.h>
#include <netinet/in.h>
#include <net/if.h>
#include <pthread.h>
#include <sched.h>
#include <semaphore.h>
#include <signal.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

typedef struct {
  const char * re_mac;
  const char * rec_mac;
  const char * rec_if;
  int recv_affinity;
  int send_affinity;
  int prepare_affinity;
  int decompress_affinity;
  int statistic_affinity;
  int ecpri_period;
  int flow_id;
  int sample_rate;
} TRXEcpriState;
static void log_error(const char * section, const char * msg, ...) {
  time_t t;
  struct tm ts;
	char line[256];
  va_list arglist;

  time(&t);
  ts = *localtime(&t);
  strftime(line, 80, "%m-%d %H:%M:%S", &ts);
  sprintf(line + strlen(line), " ERROR [%s] ", section);
  va_start(arglist, msg);
  vsprintf(line + strlen(line), msg, arglist);
  va_end(arglist);
  puts(line);
  exit(EXIT_FAILURE);
}

static void log_info(const char * section, const char * msg, ...) {
  time_t t;
  struct tm ts;
	char line[256];
  va_list arglist;

  time(&t);
  ts = *localtime(&t);
  strftime(line, 80, "%m-%d %H:%M:%S", &ts);
  sprintf(line + strlen(line), " INFO [%s] ", section);
  va_start(arglist, msg);
  vsprintf(line + strlen(line), msg, arglist);
  va_end(arglist);
  puts(line);
}
int startdpdk(TRXEcpriState * s);

int main(int argc, char * argv[]) {
  (void) argc;
  (void) argv;
  TRXEcpriState *s;

  s = malloc(sizeof(TRXEcpriState));
  memset(s, 0, sizeof(*s));

#if 0
  s->rec_mac = "80:fa:5b:92:39:c3";
  s->re_mac = "00:e0:4c:90:20:d3";
  s->rec_if = "enp53s0";
#else
  s->rec_mac = "b8:ce:f6:4b:00:22";
  s->re_mac = "b8:ce:f6:4b:00:23";
  s->rec_if = "ens5f0np0";
#endif
  s->recv_affinity = 39;
  s->send_affinity = 38;
  s->prepare_affinity = 37;
  s->decompress_affinity = 36;
  s->statistic_affinity = 35;
  s->ecpri_period = 800;
  s->flow_id = 0;
  s->sample_rate = 122880000;

  log_info("TEST-DPDK-ECPRI", "Starting test...\n");
  log_info("TEST-DPDK-ECPRI", "rec-mac: %s, re-mac: %s, rec-if: %s", s->rec_mac, s->re_mac, s->rec_if);

  startdpdk(s);

  for(int i = 0; i < 1000; i++) {
    sleep(1);
  }

}
